#!/usr/bin/python

import gnome, gnome.ui, gtk
from string import uppercase, lowercase
from types import MethodType, FunctionType
from licenses import *

FEEDBACK_INFO = 0
FEEDBACK_WARNING = 1
FEEDBACK_ERROR = 2

def _decamel(id):
  """Convert camelCaseName to Better Looking Name"""

  newid = ''

  for i in range(len(id)):
    x = id[i]
    if len(newid) == 0 or newid[-1] == ' ':
      if x != ' ' and x != '_':
        newid += x.upper()
    else:
      if x == ' ' or x == '_':
        newid += ' '
      else:
        if x in uppercase and (newid[-1] not in uppercase or (i+1 < len(id) and id[i+1] in lowercase)):
          newid += ' ' + x
        else:
          newid += x

  return newid


class Egg(list):
  def __new__(cls, target, radiogroup=None, *a, **k):
    if type(target) != FunctionType:
      return list.__new__(Egg)
    if target.func_code.co_argcount == 0:
      return list.__new__(FunctionEgg)
    if target.func_code.co_varnames[0] != "self":
      if target.func_code.co_argcount == 1:
        if target.func_code.co_varnames[0] == "filename":
          return list.__new__(FilenameFunctionEgg)
        if target.func_code.co_varnames[0] == "active":
          if radiogroup:
            return list.__new__(RadioFunctionEgg)
          else:
            return list.__new__(CheckFunctionEgg)
    else:
      if target.func_code.co_argcount == 1:
        return list.__new__(MethodEgg)
      if target.func_code.co_argcount == 2:
        if target.func_code.co_varnames[1] == "filename":
          return list.__new__(FilenameMethodEgg)
        if target.func_code.co_varnames[1] == "active":
          if radiogroup:
            return list.__new__(RadioMethodEgg)
          else:
            return list.__new__(CheckMethodEgg)

    raise TypeError('Problem with eggified method: needs to have either zero or one arguments (excluding "self")')

  widget = None
  toolitem = None
  menuitem = None
  owner = None
  target = None

  def __init__(self, target, radiogroup=None, label=None, accel=None, accelmask=gtk.gdk.CONTROL_MASK, active=False, doc=None, context=None):
    list.__init__(self)
    self.target = target
    self.accel = accel
    self.accelmask = accelmask
    self.active = active
    self.radiogroup = radiogroup
    self.context = context
    if doc:
      self.__doc__ = doc
    else:
      self.__doc__ = target.__doc__

    try:
      self.fname = _decamel(self.target.__name__)
    except:
      self.fname = "anon"

    self.label = None
    if label:
      self.label = self.baselabel = label
      if label.find("_") == -1:
        self.label = "_" + label
    else:
      self.label = self.baselabel = self.fname
      self.label = "_" + self.label

    self.stock_id = "gtk-" + self.fname.replace(" ", "-").lower()
    if self.stock_id not in gtk.stock_list_ids():
      self.stock_id = None

  def _enable(self, enabled):
    for listener in self:
      listener(enabled)

  def __get__(self, owner, ownerclass):
    self.owner = owner
    return self

  def __getattr__(self, name):
    return self.target.__getattribute__(name)

  def append(self, listener):
    list.append(self, listener)
    listener(bool(self.target))

  def __set__(self, object, value):
    state = bool(value)
    if self.widget: self.widget.set_active(state)
    if self.toolitem: self.toolitem.set_active(state)
    if self.menuitem: self.menuitem.set_active(state)
    self._enable(state)
    self.target = value

  def __call__(self, *args, **kwargs):
    return self.target(*args, **kwargs)

  def __str__(self):
    return "<%s of %s>" % (type(self).__name__, self.target.__str__())

  def getwidget(self):
    if self.widget: return self.widget
    self.widget = self._createwidget()
    _tooltips.set_tip(self.widget, self.__doc__)
    if self.context != None:
      self.context.append(self.widget.set_sensitive)
    return self.widget

  def getmenuitem(self):
    if self.menuitem: return self.menuitem
    self.menuitem = self._createmenuitem()
    if self.accel is not None:
      self.menuitem.add_accelerator("activate", _accel_group,
                                    self.accel, self.accelmask, gtk.ACCEL_VISIBLE)
    _tooltips.set_tip(self.menuitem, self.__doc__)
    if self.context != None:
      self.context.append(self.menuitem.set_sensitive)
    return self.menuitem

  def gettoolitem(self):
    if self.toolitem: return self.toolitem
    self.toolitem = self._createtoolitem()
    self.toolitem.set_label(self.label)
    self.toolitem.set_use_underline(True)
    self.toolitem.set_tooltip(_tooltips, self.__doc__)
    if self.context != None:
      self.context.append(self.toolitem.set_sensitive)
    return self.toolitem

  def _connector(self, widget):
    state = widget.get_active()
    if self.widget and widget != self.widget: self.widget.set_active(state)
    if self.toolitem and widget != self.toolitem: self.toolitem.set_active(state)
    if self.menuitem and widget != self.menuitem: self.menuitem.set_active(state)
    self._enable(state)
    self.target = state

  def _createmenuitem(self):
    menuitem = gtk.CheckMenuItem(self.label)
    menuitem.connect("toggled", self._connector)
    menuitem.set_active(self.active)
    return menuitem

  def _createtoolitem(self):
    toolitem = gtk.ToggleToolButton(stock_id=self.stock_id)
    toolitem.connect("toggled", self._connector)
    toolitem.set_active(self.active)
    return toolitem


class FunctionEgg(Egg):
  def _connector(self, widget):
    return self()

  def _createwidget(self):
    if self.stock_id:
      widget = gtk.Button(stock=self.stock_id)
    else:
      widget = gtk.Button(self.label)
    widget.connect("clicked", self._connector)
    return widget

  def _createmenuitem(self):
    if self.stock_id:
      menuitem = gtk.ImageMenuItem(stock_id=self.stock_id, accel_group=_accel_group)
    else:
      menuitem = gtk.MenuItem(self.label)
    menuitem.connect("activate", self._connector)
    return menuitem

  def _createtoolitem(self):
    toolitem = gtk.ToolButton(self.stock_id)
    toolitem.connect("clicked", self._connector)
    return toolitem

class MethodEgg(FunctionEgg):
  def __get__(self, owner, ownerclass=None):
    self.owner = owner
    return MethodType(self, owner, ownerclass)

  def _connector(self, widget):
    return self(self.owner)


class FilenameMethodEgg(MethodEgg):
  def __init__(self, target, label=None, *a, **k):
    MethodEgg.__init__(self, target, label=label, *a, **k)
    if not label:
      self.label += "..."

  def _connector(self, widget):
    searchablelabel = self.baselabel.lower()
    if searchablelabel.find("save") >= 0 or searchablelabel.find("export") >= 0:
      fileaction = gtk.FILE_CHOOSER_ACTION_SAVE
      filebutton = gtk.STOCK_SAVE
    else:
      fileaction = gtk.FILE_CHOOSER_ACTION_OPEN
      filebutton = gtk.STOCK_OPEN
    fcd = gtk.FileChooserDialog(title=self.baselabel,action=fileaction,
                                buttons=(gtk.STOCK_CANCEL,gtk.RESPONSE_CANCEL,filebutton,gtk.RESPONSE_OK))

#     preview = gtk.Image()
#     fcd.set_preview_widget(preview)
#
#     def update_preview_cb(file_chooser, preview):
#       filename = file_chooser.get_preview_filename()
#       try:
#         pixbuf = gtk.gdk.pixbuf_new_from_file_at_size(filename, 128, 128)
#         preview.set_from_pixbuf(pixbuf)
#         have_preview = True
#       except:
#         have_preview = False
#       file_chooser.set_preview_widget_active(have_preview)
#       return
#
#     fcd.connect("update-preview", update_preview_cb, preview)

    response = fcd.run()
    fcd.hide()
    if response == gtk.RESPONSE_OK:
      return self(self.owner, fcd.get_filename())

  def _createwidget(self):
    raise NotImplementedError("FilenameMethodEgg.getwidget() not yet implemented")

  def _createmenuitem(self):
    if self.stock_id:
      menuitem = gtk.ImageMenuItem(stock_id=self.stock_id, accel_group=_accel_group)
      menuitem.child.set_label(menuitem.child.get_label() + "...")
    else:
      menuitem = gtk.MenuItem(self.label)
    menuitem.connect("activate", self._connector)
    return menuitem


class CheckFunctionEgg(FunctionEgg):
  def append(self, listener):
    list.append(self, listener)
    listener(self.active)

  def _set_others_active(self, widget, state):
    if self.widget and widget != self.widget: self.widget.set_active(state)
    if self.toolitem and widget != self.toolitem: self.toolitem.set_active(state)
    if self.menuitem and widget != self.menuitem: self.menuitem.set_active(state)

  def _connector(self, widget):
    state = widget.get_active()
    self._set_others_active(widget, state)
    self._enable(state)
    return self(state)

  def _connector2(self, widget):
    state = widget.get_active()
    self._set_others_active(widget, state)

  def _createwidget(self):
    widget = gtk.CheckButton(self.label)
    if not self.toolitem and not self.menuitem:
      widget.connect("toggled", self._connector)
    else:
      widget.connect("toggled", self._connector2)
    widget.set_active(self.active)
    return widget

  def _createmenuitem(self):
    menuitem = gtk.CheckMenuItem(self.label)
    if not self.toolitem and not self.widget:
      menuitem.connect("toggled", self._connector)
    else:
      menuitem.connect("toggled", self._connector2)
    menuitem.set_active(self.active)
    return menuitem

  def _createtoolitem(self):
    toolitem = gtk.ToggleToolButton(stock_id=self.stock_id)
    if not self.menuitem and not self.widget:
      toolitem.connect("toggled", self._connector)
    else:
      toolitem.connect("toggled", self._connector2)
    toolitem.set_active(self.active)
    return toolitem

class CheckMethodEgg(MethodEgg, CheckFunctionEgg):
  def _connector(self, widget):
    state = widget.get_active()
    self._set_others_active(widget, state)
    self._enable(state)
    return self(self.owner, state)


class RadioMethodEgg(CheckMethodEgg):
  def __init__(self, *a, **k):
    CheckMethodEgg.__init__(self, *a, **k)
    try: RadioMethodEgg.__radiogroups
    except: RadioMethodEgg.__radiogroups = {}
    try: RadioMethodEgg.__radiomenugroups
    except: RadioMethodEgg.__radiomenugroups = {}
    try: RadioMethodEgg.__radiotoolgroups
    except: RadioMethodEgg.__radiotoolgroups = {}

  def _createwidget(self):
    try: widgetgroup = RadioMethodEgg.__radiogroups[self.radiogroup]
    except: widgetgroup = None
    widget = gtk.RadioButton(group=widgetgroup, label=self.label)
    if widgetgroup==None: RadioMethodEgg.__radiogroups[self.radiogroup] = widget
    if not self.toolitem and not self.menuitem:
      widget.connect("toggled", self._connector)
    else:
      widget.connect("toggled", self._connector2)
    widget.set_active(self.active)
    return widget

  def _createmenuitem(self):
    try: menugroup = RadioMethodEgg.__radiomenugroups[self.radiogroup]
    except: menugroup = None
    menuitem = gtk.RadioMenuItem(group=menugroup, label=self.label)
    if menugroup==None: RadioMethodEgg.__radiomenugroups[self.radiogroup] = menuitem
    if not self.toolitem and not self.widget:
      menuitem.connect("toggled", self._connector)
    else:
      menuitem.connect("toggled", self._connector2)
    menuitem.set_active(self.active)
    return menuitem

  def _createtoolitem(self):
    try: toolgroup = RadioMethodEgg.__radiotoolgroups[self.radiogroup]
    except: toolgroup = None
    toolitem = gtk.RadioToolButton(group=toolgroup, stock_id=self.stock_id)
    if toolgroup==None: RadioMethodEgg.__radiotoolgroups[self.radiogroup] = toolitem
    if not self.menuitem and not self.widget:
      toolitem.connect("toggled", self._connector)
    else:
      toolitem.connect("toggled", self._connector2)
    toolitem.set_active(self.active)
    return toolitem

class _EggTargetNone(object):
  pass

_egg_target_none = _EggTargetNone()

def eggify(target=_egg_target_none, *a, **k):
  """Decorator to add a getwidget function to a method.
     
     Example:
     import egg
     
     class MyApp(egg.App):
       @egg.eggify()
       def clickMe(self):
         print "You pressed my button!"
     
       def __init__(self):
         egg.App.__init__(self)
         self.set_contents(self.clickMe.getwidget())
     
     if __name__ == "__main__":
       a=MyApp()
       a.run()
  """
  if target != _egg_target_none:
    return Egg(target, *a, **k)
  else:
    return lambda target: Egg(target=target, *a, **k)


class EggWidget(object):
  def getwidget(self):
    return self


def _menugen(menu, itemlist):
  """Helper function for recursively generating menus"""
  for item in itemlist:
    if item is None:
      menu.append(gtk.SeparatorMenuItem())
    elif type(item) is list:
      if item[0].find("_") == -1:
        item[0] = "_" + item[0]
      menuitem = gtk.MenuItem(item[0])
      submenu = gtk.Menu()
      _menugen(submenu, item[1:])
      menuitem.set_submenu(submenu)
      menu.append(menuitem)
    elif callable(item):
      menu.append(item.getmenuitem())

class Menu(gtk.MenuBar, EggWidget):
  """Single Menu class for creating and manipulating the whole menubar"""
  def __init__(self, startvals):
    gtk.MenuBar.__init__(self)
    _menugen(self, startvals)
    self.show_all()

class Label(gtk.Label, EggWidget):
  def __init__(self, label):
    gtk.Label.__init__(self, label)

class VBox(gtk.VBox, EggWidget):
  def __init__(self, *args, **kwargs):
    gtk.VBox.__init__(self)
    if type(args[0]) is str:
      frame = gtk.Frame(label=args[0])
      frame.add(VBox(*args[1:]))
      self.add(frame)
      return
    for arg in args:
      try:
        self.add(arg.getwidget())
      except AttributeError:
        try:
          arg.__iter__
          self.add(HBox(*arg))
        except AttributeError:
          self.add(Label(str(arg)))

class HBox(gtk.HButtonBox, EggWidget):
  def __init__(self, *args, **kwargs):
    gtk.HButtonBox.__init__(self)
    for arg in args:
      try:
        self.add(arg.getwidget())
      except AttributeError:
        try:
          arg.__iter__
          self.add(VBox(*arg))
        except AttributeError:
          self.add(Label(str(arg)))

class VButtonBox(gtk.VButtonBox, EggWidget):
  def __init__(self, items=None):
    gtk.VButtonBox.__init__(self)
    if items:
      for item in items:
        self.add(item.getwidget())

class HButtonBox(gtk.HButtonBox, EggWidget):
  def __init__(self, items=None):
    gtk.HButtonBox.__init__(self)
    if items:
      for item in items:
        self.add(item.getwidget())

class Window(gtk.Window, EggWidget):
  def __init__(self, *args, **kwargs):
    gtk.Window.__init__(self)
    if kwargs.has_key("title"):
      title = kwargs["title"]
    else:
      title = _decamel(self.__class__.__name__)
    self.add(VBox(*args))
    self.set_title(title)

  def __call__(self):
    self.show_all()

def mkwidget(*args, **kwargs):
  vbox = VBox(args)

def _gen_callback_bound(f):
  """Turns a bound method into a pygtk callback."""
  def cb(widget):
    return f()
  return cb

def connect(widget, name, cb):
  """Connects the named signal of widget with the specified callback.
     
     The callback should not accept any arguments.  That is, it should
     be a function or a bound method.
  """
  # This should be already converted to a gtk callback with a decorator, else
  # it is assumed to be a function or a bound method, and connected as such.
  try: connector = cb.connector
  except: connector = _gen_callback_bound(cb)
  return gnome.ui.App.connect(widget, name, connector)

class About(gtk.AboutDialog, EggWidget):
  """Instantiate and call this class to provide the user with information about the program."""

  def __init__(self, name=None, version=None, comments=None, copyright=None,
               license=None, wrap_license=None, website=None,
               website_label=None, authors=None, documenters=None,
               artists=None, translator_credits=None, logo=None, icon_name=None
              ):
    gtk.AboutDialog.__init__(self)
    if name: self.set_name(name)
    if version: self.set_version(version)
    if comments: self.set_comments(comments)
    if copyright: self.set_copyright(copyright)
    if license: self.set_license(license)
    if wrap_license: self.set_wrap_license(wrap_license)
    if website: self.set_website(website)
    if authors: self.set_authors(authors)
    if documenters: self.set_documenters(documenters)
    if artists: self.set_artists(artists)
    if translator_credits: self.set_translator_credits(translator_credits)
    if logo: self.set_logo(logo)
    if icon_name: self.set_icon_name(icon_name)

  def __call__(self):
    self.run()
    self.hide()

class Toolbar(gtk.Toolbar, EggWidget):
  def __init__(self, toolbaritems):
    gtk.Toolbar.__init__(self)
    for item in toolbaritems:
      if item is None:
        self.add(gtk.SeparatorToolItem())
      else:
        self.add(item.gettoolitem())
    self.show_all()

_accel_group = gtk.AccelGroup()
_tooltips = gtk.Tooltips()

class App(gnome.ui.App, EggWidget):
  _the_App = None
  _init_called = False

  def feedback(self, data, type=FEEDBACK_INFO):
    if type == FEEDBACK_INFO:
      self.statusbar.set_status(str(data))

  @eggify()
  def quit(self):
    gtk.main_quit()

  @eggify()
  def about(self):
    self._about()

  # Ensure there's only one App
  def __new__(cls, *p, **k):
    if not App._the_App:
      App._the_App = gnome.ui.App.__new__(cls, *p, **k)
    return App._the_App

  def __init__(self, name=None, windowtitle=None, version="0.0", menu=None,
               tools=None,
               copyright=None, comments=None, license=None, wrap_license=None,
               website=None, website_label=None, authors=None,
               documenters=None, artists=None, translator_credits=None,
               logo=None, icon_name=None, size=(400,400)
              ):
    if self._init_called:
      print "Warning: Not creating second App"
      return
    self._init_called = True
    if not name: name = self.__class__.__name__
    gnome.init(name, version)
    if not windowtitle: windowtitle = _decamel(name)
    self.quit.im_func.__doc__ = "Quit %s" % windowtitle
    self.about.im_func.__doc__ = "View information about %s" % windowtitle
    gnome.ui.App.__init__(self, name, windowtitle)
    self.add_accel_group(_accel_group)
    self.statusbar = gnome.ui.AppBar(True, True, gnome.ui.PREFERENCES_USER)
    if not comments: comments=self.__doc__
    self._about = About(name=windowtitle, version=version, comments=comments,
                        copyright=copyright, license=license,
                        wrap_license=wrap_license, website=website,
                        website_label=website_label, authors=authors,
                        documenters=documenters, artists=artists,
                        translator_credits=translator_credits, logo=logo,
                        icon_name=icon_name)
    if menu == None:
      menu = [["_File", self.quit], ["_Help", self.about]]
    mymenubar = Menu(menu)
    self.set_menus(mymenubar)
    if tools == None:
      tools = [self.quit]
    mytoolbar = Toolbar(tools)
    self.set_toolbar(mytoolbar)
    self.set_statusbar(self.statusbar)
    self.set_size_request(size[0], size[1])
    connect(self, 'destroy', self.quit)

  def __call__(self):
    self.show_all()
    gtk.main()


def feedback(*p, **k):
  return App._the_App.feedback(*p, **k)


if __name__ == "__main__":
  app=App()
  app()
